# import pandas as pd
#
# from visions.relations import InferenceRelation
# from visions.relations.relations_utils import values_are_consecutive
# from visions.types.ordinal import to_ordinal
#
#
# def is_ordinal_int(s: pd.Series, state: dict) -> bool:
#     initial_element = 1
#     distinct_values = list(s.unique())
#     return (
#         initial_element in distinct_values
#         and values_are_consecutive(distinct_values)
#         and 2 < len(distinct_values) < 10
#     )
#
#
# def integer_to_ordinal() -> InferenceRelation:
#     from visions.types import Integer
#
#     return InferenceRelation(
#         Integer, relationship=is_ordinal_int, transformer=to_ordinal
#     )



import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas import test_utils
from visions.backends.pandas.series_utils import series_not_empty, series_not_sparse
from visions.backends.pandas.types.float import string_is_float
from visions.backends.shared.parallelization_engines import pandas_apply
from visions.types.complex import Complex
from visions.types.string import String


def imaginary_in_string(series: pd.Series, imaginary_indicator: tuple = ("j", "i")):
    return any(any(v in s for v in imaginary_indicator) for s in series)


@Complex.register_relationship(String, pd.Series)
def string_is_complex(series: pd.Series, state: dict) -> bool:
    def f(s: pd.Series) -> pd.Series:
        return pandas_apply(s, complex)

    coerced_series = test_utils.option_coercion_evaluator(f)(series)
    return (
        coerced_series is not None
        and not string_is_float(series, state)
        and imaginary_in_string(series)
    )


@Complex.register_transformer(String, pd.Series)
def string_to_complex(series: pd.Series, state: dict) -> pd.Series:
    return pandas_apply(series, complex)


@Complex.contains_op.register
@series_not_sparse
@series_not_empty
def complex_contains(series: pd.Series, state: dict) -> bool:
    return pdt.is_complex_dtype(series)



import datetime
from typing import Dict, Iterable

import numpy as np
import pandas as pd

from visions.backends.pandas.test_utils import pandas_version
from visions.backends.pandas.types.boolean import hasnan_bool_name


def get_sequences() -> Dict[str, Iterable]:
    sequences = {
        "float_series6": pd.Series([np.nan, 1.1], dtype=np.single),
        "bool_series2": pd.Series([True, False, False, True], dtype=bool),
        "nullable_bool_series": pd.Series([True, False, None], dtype=hasnan_bool_name),
        "int_str_range": pd.Series(range(20)).astype("str"),
        "Int64_int_series": pd.Series([1, 2, 3], dtype="Int64"),
        "Int64_int_nan_series": pd.Series([1, 2, 3, np.nan], dtype="Int64"),
        "pd_uint32": pd.Series([1, 2, 3, 4], dtype="UInt32"),
        "categorical_int_series": pd.Series([1, 2, 3], dtype="category"),
        "categorical_char": pd.Series(
            pd.Categorical(
                ["A", "B", "C", "C", "B", "A"],
                categories=["A", "B", "C"],
                ordered=False,
            ),
        ),
        "categorical_float_series": pd.Series([1.0, 2.0, 3.1], dtype="category"),
        "categorical_string_series": pd.Series(["Georgia", "Sam"], dtype="category"),
        "categorical_complex_series": pd.Series(
            [complex(0, 0), complex(1, 2), complex(3, -1)],
            dtype="category",
        ),
        "ordinal": pd.Series(
            pd.Categorical(
                ["A", "B", "C", "C", "B", "A"],
                categories=["A", "B", "C"],
                ordered=True,
            ),
        ),
        "timestamp_series": pd.to_datetime(
            pd.Series(
                [
                    datetime.datetime(2017, 3, 5, 12, 2),
                    datetime.datetime(2019, 12, 4),
                ],
            )
        ),
        "timestamp_series_nat": pd.to_datetime(
            pd.Series(
                [
                    datetime.datetime(2017, 3, 5),
                    datetime.datetime(2019, 12, 4, 3, 2, 0),
                    pd.NaT,
                ],
            )
        ),
        "date_series_nat": pd.to_datetime(
            pd.Series(
                [
                    datetime.datetime(2017, 3, 5),
                    datetime.datetime(2019, 12, 4),
                    pd.NaT,
                ],
            )
        ),
        "timestamp_aware_series": pd.Series(
            pd.date_range(
                start="2013-05-18 12:00:01",
                periods=2,
                freq="H",
                tz="Europe/Brussels",
            )
        ),
        "datetime": pd.to_datetime(
            pd.Series(
                [
                    datetime.date(2011, 1, 1),
                    datetime.date(2012, 1, 2),
                    datetime.date(2013, 1, 1),
                ],
            )
        ),
        # http://pandas-docs.github.io/pandas-docs-travis/user_guide/timeseries.html#timestamp-limitations
        # pd.to_datetime(
        #     pd.Series(
        #         [
        #             datetime.datetime(year=1, month=1, day=1, hour=8, minute=43, second=12),
        #             datetime.datetime(year=1, month=1, day=1, hour=9, minute=43, second=12),
        #             datetime.datetime(
        #                 year=1, month=1, day=1, hour=10, minute=43, second=12
        #             ),
        #         ],
        #         name="datetime_to_time",
        #     )
        # ),
        "timedelta_series": pd.Series([pd.Timedelta(days=i) for i in range(3)]),
        "timedelta_series_nat": pd.Series(
            [pd.Timedelta(days=i) for i in range(3)] + [pd.NaT],
        ),
        "timedelta_negative": pd.Series(
            [
                pd.Timedelta("1 days 00:03:43"),
                pd.Timedelta("5 days 12:33:57"),
                pd.Timedelta("0 days 01:25:07"),
                pd.Timedelta("-2 days 13:46:56"),
                pd.Timedelta("1 days 23:49:25"),
            ],
        ),
        "empty_float": pd.Series([], dtype=float),
        "empty_int64": pd.Series([], dtype="Int64"),
        "empty_object": pd.Series([], dtype="object"),
        "empty_bool": pd.Series([], dtype=bool),
        "float_series4": pd.Series([1, 2, 3.05, 4], dtype=np.float64),
        # Null Sequences
        "all_null_none": pd.Series([None, None]),
        "all_null_nan": pd.Series([np.nan, np.nan]),
        "all_null_nat": pd.Series([pd.NaT, pd.NaT]),
        "all_null_empty_str": pd.Series(["", ""]),
    }

    if pandas_version[0] >= 1:
        sequences["string_dtype_series"] = pd.Series(
            ["Patty", "Valentine"], dtype="string"
        )

    return sequences



from datetime import datetime

import numpy as np

from visions.backends.numpy.array_utils import (
    all_type,
    array_handle_nulls,
    array_not_empty,
)
from visions.types.object import Object


def not_excluded_type(array: np.ndarray, excludes) -> bool:

    if len(array) == 0 or not isinstance(array[0], excludes):
        return True

    dtype = type(array[0])
    return not all_type(array, dtype)


@Object.contains_op.register
@array_handle_nulls
@array_not_empty
def object_contains(array: np.ndarray, state: dict) -> bool:
    if np.issubdtype(array.dtype, np.str_):
        return True

    if not np.issubdtype(array.dtype, np.object_):
        return False

    return not_excluded_type(array, (bool, int, datetime))



import functools
from typing import Callable, Sequence, Tuple, TypeVar, Union

import numpy as np

from visions.backends.shared.nan_handling import nan_mask
from visions.backends.shared.utilities import has_import

has_numba = has_import("numba")

if has_numba:
    import numba as nb

T = TypeVar("T")


def array_handle_nulls(fn: Callable[..., bool]) -> Callable[..., bool]:
    """Decorator for nullable arrays"""

    handles_missing = array_not_empty(fn)

    @functools.wraps(fn)
    def inner(array: np.ndarray, *args, **kwargs) -> bool:
        array = array[nan_mask(array)]
        return handles_missing(array, *args, **kwargs)

    return inner


def array_not_empty(fn: Callable[..., bool]) -> Callable[..., bool]:
    """Decorator to exclude empty arrays"""

    @functools.wraps(fn)
    def inner(array: np.ndarray, *args, **kwargs) -> bool:
        if array.shape[0] == 0:
            return False
        return fn(array, *args, **kwargs)

    return inner


def _base_all_type(array: np.ndarray, dtypes: Union[type, Tuple[type, ...]]) -> bool:
    return all(isinstance(v, dtypes) for v in array)


if has_numba:
    # TODO: This only works when the numpy array dtype falls under a few categories
    # There are alternative implementations with forceobj=True which work in all cases
    # including the use of isinstance, but in those cases worst case performance can be substantially worse
    # than the default python implementation.
    def all_type_numba(dtype: Union[Tuple, T]):
        @nb.jit(nopython=True)
        def inner(array: np.ndarray) -> bool:
            for i in nb.prange(array.size):
                if type(array[i]) is not dtype:
                    return False
            return True

        return inner

    def all_type(array: np.ndarray, dtypes: Union[type, Tuple[type, ...]]) -> bool:
        return _base_all_type(array, dtypes)


else:

    def all_type(array: np.ndarray, dtypes: Union[type, Tuple[type, ...]]) -> bool:
        return _base_all_type(array, dtypes)



from typing import Sequence

from visions.types.email_address import FQDA, EmailAddress, _to_email
from visions.types.string import String


@EmailAddress.register_relationship(String, Sequence)
def string_is_email(sequence: Sequence, state: dict) -> bool:
    try:
        return all(
            value.local and value.fqdn for value in string_to_email(sequence, state)
        )
    except (ValueError, TypeError, AttributeError):
        return False


@EmailAddress.register_transformer(String, Sequence)
def string_to_email(sequence: Sequence, state: dict) -> Sequence:
    return tuple(map(_to_email, sequence))


@EmailAddress.contains_op.register
def email_address_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(value, FQDA) for value in sequence)



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Optional, Sequence, Type, Union, cast

import attr
from multimethod import multimethod

from visions.relations import TypeRelation

_DEFAULT = object()


class RelationsIterManager:
    """Class to enable to treat relations as dict"""

    def __init__(self, relations: Sequence[TypeRelation]):
        self._keys: Dict["Type[VisionsBaseType]", int] = {
            item.related_type: i for i, item in enumerate(relations)
        }
        self.values = tuple(relations)

    def __getitem__(self, index: Union["Type[VisionsBaseType]", int]) -> TypeRelation:
        idx = index if isinstance(index, int) else self._keys[index]
        return self.values[idx]

    def get(
        self, index: Union["Type[VisionsBaseType]", int], default: Any = _DEFAULT
    ) -> Union[TypeRelation, Any]:
        try:
            return self[index]
        except (IndexError, KeyError) as err:
            if default is _DEFAULT:
                raise err
            else:
                return default

    def __iter__(self):
        yield from self.values


class VisionsBaseTypeMeta(ABCMeta):
    _relations: Optional[RelationsIterManager] = None

    def __contains__(cls, sequence: Sequence) -> bool:
        return cls.contains_op(sequence, dict())

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        raise NotImplementedError

    @staticmethod
    def contains_op(item: Any, state: dict) -> bool:
        raise NotImplementedError

    @property
    def relations(cls) -> RelationsIterManager:
        from visions.relations.relations import IdentityRelation

        if cls._relations is None:
            cls._relations = RelationsIterManager(
                [
                    attr.evolve(
                        r,
                        type=cls,
                        relationship=cls.contains_op
                        if r.relationship is None
                        else r.relationship,
                    )
                    if isinstance(r, IdentityRelation)
                    else attr.evolve(
                        r,
                        type=cls,
                        relationship=multimethod(r.relationship)
                        if r.relationship is not None
                        else None,
                        transformer=multimethod(r.transformer),
                    )
                    for r in cls.get_relations()
                ]
            )
        return cls._relations

    def __add__(cls, other):
        from visions.types import Generic
        from visions.typesets import VisionsTypeset

        if not any(issubclass(x, Generic) for x in [cls, other]):
            return VisionsTypeset({Generic, cls, other})
        return VisionsTypeset({cls, other})

    def __str__(cls) -> str:
        return str(cls.__name__)

    def __repr__(cls) -> str:
        return str(cls)


class VisionsBaseType(metaclass=VisionsBaseTypeMeta):
    """Abstract implementation of a vision type.

    Provides a common API for building custom visions data types.
    """

    def __init__(self):
        pass

    @staticmethod
    @abstractmethod
    def get_relations() -> Sequence[TypeRelation]:
        raise NotImplementedError

    @classmethod
    def register_transformer(
        cls, relation: "Type[VisionsBaseType]", dispatch_type: Any
    ):
        relation_transformer = cls.relations[relation].transformer
        return cast(Any, relation_transformer).register(dispatch_type, dict)

    @classmethod
    def register_relationship(
        cls, relation: "Type[VisionsBaseType]", dispatch_type: Any
    ):
        relation_relationship = cls.relations[relation].relationship
        return cast(Any, relation_relationship).register(dispatch_type, dict)

    @staticmethod
    @multimethod
    @abstractmethod
    def contains_op(sequence: Any, state: Any) -> bool:
        raise NotImplementedError



import pandas as pd
from pandas.api import types as pdt

from visions.backends.pandas.series_utils import (
    series_handle_nulls,
    series_not_empty,
    series_not_sparse,
)
from visions.types.object import Object

pandas_has_string_dtype_flag = hasattr(pdt, "is_string_dtype")


@Object.contains_op.register
@series_not_sparse
@series_handle_nulls
@series_not_empty
def object_contains(series: pd.Series, state: dict) -> bool:
    is_object = pdt.is_object_dtype(series)
    if is_object:
        ret = True
    elif pandas_has_string_dtype_flag:
        ret = pdt.is_string_dtype(series) and not pdt.is_categorical_dtype(series)
    else:
        ret = False
    return ret



import visions.backends.spark.types.boolean
import visions.backends.spark.types.categorical
import visions.backends.spark.types.date
import visions.backends.spark.types.float
import visions.backends.spark.types.integer
import visions.backends.spark.types.numeric
import visions.backends.spark.types.object
import visions.backends.spark.types.string



import visions.backends.pandas.types.boolean
import visions.backends.pandas.types.categorical
import visions.backends.pandas.types.complex
import visions.backends.pandas.types.count
import visions.backends.pandas.types.date
import visions.backends.pandas.types.date_time
import visions.backends.pandas.types.email_address
import visions.backends.pandas.types.file
import visions.backends.pandas.types.float
import visions.backends.pandas.types.geometry
import visions.backends.pandas.types.image
import visions.backends.pandas.types.integer
import visions.backends.pandas.types.ip_address
import visions.backends.pandas.types.numeric
import visions.backends.pandas.types.object
import visions.backends.pandas.types.ordinal
import visions.backends.pandas.types.path
import visions.backends.pandas.types.sparse
import visions.backends.pandas.types.string
import visions.backends.pandas.types.time
import visions.backends.pandas.types.time_delta
import visions.backends.pandas.types.url
import visions.backends.pandas.types.uuid



import datetime
import os
import uuid
from ipaddress import IPv4Address, IPv6Address
from pathlib import Path, PurePosixPath, PureWindowsPath
from typing import Dict, Sequence, cast
from urllib.parse import urlparse

from visions.types.email_address import FQDA


def get_sequences() -> Dict[str, Sequence]:
    base_path = Path(__file__).parent.parent.parent.absolute()

    sequences = {
        "int_series": [1, 2, 3],
        "int_range": range(10),
        "int_series_boolean": [1, 0, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 0],
        "float_series": [1.0, 2.1, 3.0],
        "float_series2": [1.0, 2.0, 3.0, 4.0],
        "string_series": ["Patty", "Valentine"],
        "string_unicode_series": ["mack", "the", "finger"],
        "string_num": ["1.0", "2.0", "3.0"],
        "string_flt": ["1.0", "45.67", "3.5"],
        "string_bool_nan": ["True", "False", None],
        "str_url": [
            "http://www.cwi.nl:80/%7Eguido/Python.html",
            "https://github.com/dylan-profiling/hurricane",
        ],
        "path_series_windows_str": [
            r"C:\\home\\user\\file.txt",
            r"C:\\home\\user\\test2.txt",
        ],
        "path_series_linux_str": [r"/home/user/file.txt", r"/home/user/test2.txt"],
        "str_int_leading_zeros": ["0011", "12"],
        "str_float_non_leading_zeros": ["0.0", "0.04", "0"],
        "str_int_zeros": ["0.0", "0.000", "0", "2"],
        "bool_series": [True, False],
        "bool_nan_series": [True, False, None],
        "str_complex": ["(1+1j)", "(2+2j)", "(10+100j)"],
        "str_complex_nan": ["(1+1j)", "(2+2j)", "(10+100j)", "NaN"],
        "complex_series_py": [complex(0, 0), complex(1, 2), complex(3, -1)],
        "complex_series_py_float": [complex(0, 0), complex(1, 0), complex(3, 0)],
        "string_date": ["1937-05-06", "20/4/2014"],
        "timestamp_string_series": ["1941-05-24", "13/10/2016"],
        "py_datetime_str": ["1941-05-24 00:05:00", "2016-10-13 00:10:00"],
        "date": [
            datetime.date(2011, 1, 1),
            datetime.date(2012, 1, 2),
            datetime.date(2013, 1, 1),
        ],
        "time": [
            datetime.time(8, 43, 12),
            datetime.time(9, 43, 12),
            datetime.time(10, 43, 12),
        ],
        "path_series_linux": [
            PurePosixPath("/home/user/file.txt"),
            PurePosixPath("/home/user/test2.txt"),
        ],
        "path_series_linux_missing": [
            PurePosixPath("/home/user/file.txt"),
            PurePosixPath("/home/user/test2.txt"),
            None,
        ],
        "path_series_windows": [
            PureWindowsPath("C:\\home\\user\\file.txt"),
            PureWindowsPath("C:\\home\\user\\test2.txt"),
        ],
        "url_series": [
            urlparse("http://www.cwi.nl:80/%7Eguido/Python.html"),
            urlparse("https://github.com/dylan-profiling/hurricane"),
        ],
        "url_none_series": [
            urlparse("http://www.cwi.nl:80/%7Eguido/Python.html"),
            urlparse("https://github.com/dylan-profiling/hurricane"),
            None,
        ],
        "uuid_series": [
            uuid.UUID("0b8a22ca-80ad-4df5-85ac-fa49c44b7ede"),
            uuid.UUID("aaa381d6-8442-4f63-88c8-7c900e9a23c6"),
            uuid.UUID("00000000-0000-0000-0000-000000000000"),
        ],
        "uuid_series_missing": [
            uuid.UUID("0b8a22ca-80ad-4df5-85ac-fa49c44b7ede"),
            uuid.UUID("aaa381d6-8442-4f63-88c8-7c900e9a23c6"),
            uuid.UUID("00000000-0000-0000-0000-000000000000"),
            None,
        ],
        "uuid_series_str": [
            "0b8a22ca-80ad-4df5-85ac-fa49c44b7ede",
            "aaa381d6-8442-4f63-88c8-7c900e9a23c6",
            "00000000-0000-0000-0000-000000000000",
        ],
        "mixed_list[str,int]": [[1, ""], [2, "Rubin"], [3, "Carter"]],
        "mixed_dict": [
            {"why": "did you"},
            {"bring him": "in for he"},
            {"aint": "the guy"},
        ],
        "callable": [os.getcwd, os.stat, os.kill],
        "module": [os, uuid],
        "textual_float": ["1.1", "2"],
        "textual_float_nan": ["1.1", "2", "NAN"],
        "mixed_integer": ["a", 1],
        "mixed_list": [[True], [False], [False]],
        "ip_str": ["127.0.0.1", "127.0.0.1"],
        "empty": [],
        "ip": [IPv4Address("127.0.0.1"), IPv4Address("127.0.0.1")],
        "ip_missing": [IPv4Address("127.0.0.1"), None, IPv4Address("127.0.0.1")],
        "ip_mixed_v4andv6": [IPv6Address("0:0:0:0:0:0:0:1"), IPv4Address("127.0.0.1")],
        "file_test_py": [
            Path(os.path.join(base_path, "test/series.py")).absolute(),
            Path(os.path.join(base_path, "test/__init__.py")).absolute(),
            Path(os.path.join(base_path, "test/utils.py")).absolute(),
        ],
        "file_mixed_ext": [
            Path(os.path.join(base_path, "py.typed")).absolute(),
            Path(os.path.join(base_path, "test/data", "file.html")).absolute(),
            Path(os.path.join(base_path, "test/series.py")).absolute(),
        ],
        "file_test_py_missing": [
            Path(os.path.join(base_path, "test/series.py")).absolute(),
            None,
            Path(os.path.join(base_path, "test/__init__.py")).absolute(),
            None,
            Path(os.path.join(base_path, "test/utils.py")).absolute(),
        ],
        "image_png": [
            Path(
                os.path.join(
                    base_path,
                    "test/data",
                    "img.png",
                )
            ).absolute(),
            Path(
                os.path.join(
                    base_path,
                    "test/data",
                    "img.jpeg",
                )
            ).absolute(),
            Path(
                os.path.join(
                    base_path,
                    "test/data",
                    "img.jpg",
                )
            ).absolute(),
        ],
        "image_png_missing": [
            Path(
                os.path.join(
                    base_path,
                    "test/data",
                    "img.png",
                )
            ).absolute(),
            Path(
                os.path.join(
                    base_path,
                    "test/data",
                    "img.jpeg",
                )
            ).absolute(),
            None,
            Path(
                os.path.join(
                    base_path,
                    "test/data",
                    "img.jpg",
                )
            ).absolute(),
            None,
        ],
        "email_address": [FQDA("test", "example.com"), FQDA("info", "example.eu")],
        "email_address_missing": [
            FQDA("test", "example.com"),
            FQDA("info", "example.eu"),
            None,
        ],
        "email_address_str": ["test@example.com", "info@example.eu"],
    }
    assert all(isinstance(v, Sequence) for v in sequences.values())
    return cast(Dict[str, Sequence], sequences)



import pathlib

import pandas as pd

from visions.backends.pandas.series_utils import series_handle_nulls, series_not_empty
from visions.types.file import File


@File.contains_op.register
@series_not_empty
@series_handle_nulls
def file_contains(series: pd.Series, state: dict) -> bool:
    return all(isinstance(p, pathlib.Path) and p.exists() for p in series)



import numpy as np
from packaging import version

from visions.backends.numpy import test_utils
from visions.backends.numpy.array_utils import array_not_empty
from visions.backends.numpy.types.float import string_is_float
from visions.types.complex import Complex
from visions.types.string import String

_OLD_NUMPY = version.parse(np.version.version) <= version.parse("1.19.0")


def imaginary_in_string(array: np.ndarray, imaginary_indicator: tuple = ("j", "i")):
    return any(any(v in s for v in imaginary_indicator) for s in array)


@Complex.register_transformer(String, np.ndarray)
def string_to_complex(array: np.array, state: dict) -> np.ndarray:
    if _OLD_NUMPY:
        return np.array([complex(v) for v in array])
    else:
        return array.astype(complex)


@Complex.register_relationship(String, np.ndarray)
def string_is_complex(array: np.ndarray, state: dict) -> bool:
    coerced_array = test_utils.option_coercion_evaluator(
        lambda x: string_to_complex(x, state)
    )(array)
    return (
        coerced_array is not None
        and not string_is_float(array, state)
        and imaginary_in_string(array)
    )


@Complex.contains_op.register
@array_not_empty
def complex_contains(array: np.ndarray, state: dict) -> bool:
    return np.issubdtype(array.dtype, complex)



import pathlib

import pandas as pd

from visions.backends.pandas.series_utils import series_handle_nulls, series_not_empty
from visions.backends.shared.parallelization_engines import pandas_apply
from visions.types.path import Path
from visions.types.string import String


@Path.register_relationship(String, pd.Series)
def string_is_path(series: pd.Series, state: dict) -> bool:
    try:
        s = string_to_path(series.copy(), state)
        return pandas_apply(s, lambda x: x.is_absolute()).all()
    except TypeError:
        return False


@Path.register_transformer(String, pd.Series)
def string_to_path(series: pd.Series, state: dict) -> pd.Series:
    s = pandas_apply(series, pathlib.PureWindowsPath)
    if not pandas_apply(s, lambda x: x.is_absolute()).all():
        return pandas_apply(series, pathlib.PurePosixPath)
    else:
        return s


@Path.contains_op.register
@series_not_empty
@series_handle_nulls
def path_contains(series: pd.Series, state: dict) -> bool:
    return all(isinstance(x, pathlib.PurePath) and x.is_absolute() for x in series)



from typing import Dict, List, Sequence

from visions.backends.python.series_utils import (
    sequence_handle_none,
    sequence_not_empty,
)
from visions.types import Boolean, Object, String


def get_boolean_coercions(id: str) -> List[Dict]:
    coercion_map = {
        "default": [{"true": True, "false": False}],
        "en": [
            {"true": True, "false": False},
            {"y": True, "n": False},
            {"yes": True, "no": False},
        ],
        "nl": [
            {"true": True, "false": False},
            {"ja": True, "nee": False},
            {"j": True, "n": False},
        ],
    }
    return coercion_map[id]


@sequence_not_empty
@sequence_handle_none
def is_bool(sequence: Sequence, state: dict):
    return all(isinstance(value, bool) for value in sequence)


def to_bool(sequence: Sequence, state: dict):
    return map(bool, sequence)


@Boolean.register_relationship(Object, Sequence)
def object_is_bool(sequence: Sequence, state: dict) -> bool:
    return is_bool(sequence, state)


@Boolean.register_transformer(Object, Sequence)
def object_to_bool(sequence: Sequence, state: dict) -> Sequence:
    return to_bool(sequence, state)


@Boolean.register_relationship(String, Sequence)
@sequence_handle_none
def string_is_bool(sequence: Sequence, state: dict):
    return all(value.lower() in {"true", "false"} for value in sequence)


@Boolean.register_transformer(String, Sequence)
def string_to_bool(sequence: Sequence, state: dict):
    return map(lambda v: v.lower() == "true" if isinstance(v, str) else v, sequence)


@Boolean.contains_op.register
def boolean_contains(sequence: Sequence, state: dict) -> bool:
    return is_bool(sequence, state)



from datetime import time
from typing import Sequence

# from visions.types.date_time import DateTime
from visions.types.time import Time

# @Time.register_relationship(DateTime, Sequence)
# def datetime_is_time(sequence: Sequence, state: dict) -> bool:
#     value = date(1, 1, 1)
#     return all(v == value for v in sequence)
#
#
# @Time.register_transformer(DateTime, Sequence)
# def datetime_to_time(sequence: Sequence, state: dict) -> Sequence:
#     return map(lambda v: v.time(), sequence)


@Time.contains_op.register
def time_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(value, time) for value in sequence)



import numbers
from typing import Sequence

from visions.types.numeric import Numeric


@Numeric.contains_op.register
def numeric_contains_op(sequence: Sequence, state: dict):
    return all(
        isinstance(value, numbers.Number) and not isinstance(value, bool)
        for value in sequence
    )



from typing import Sequence

from visions.backends.python.series_utils import sequence_not_empty
from visions.types.complex import Complex
from visions.types.float import Float
from visions.types.string import String


def no_leading_zeros(sequence, coerced_sequence) -> bool:
    return not any(s[0] == "0" and c > 1 for s, c in zip(sequence, coerced_sequence))


@Float.register_relationship(String, Sequence)
def string_is_float(sequence: Sequence, state: dict) -> bool:
    try:
        coerced = list(string_to_float(sequence, state))
        return no_leading_zeros(sequence, coerced)
    except ValueError:
        return False


@Float.register_transformer(String, Sequence)
def string_to_float(sequence: Sequence, state: dict) -> Sequence:
    return tuple(map(float, sequence))


@Float.register_relationship(Complex, Sequence)
def complex_is_float(sequence: Sequence, state: dict) -> bool:
    try:
        return all(value.imag == 0 for value in sequence)
    except ValueError:
        return False


@Float.register_transformer(Complex, Sequence)
def complex_to_float(sequence: Sequence, state: dict) -> Sequence:
    return list(map(lambda v: v.real, sequence))


@Float.contains_op.register
@sequence_not_empty
def float_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(value, float) for value in sequence)



from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import DateType

from visions.types.date import Date


@Date.contains_op.register
def date_contains(sequence: DataFrame, state: dict) -> bool:
    if len(sequence.schema) != 1:
        return False

    dtype = sequence.schema[0].dataType
    return isinstance(dtype, DateType)



from ipaddress import _BaseAddress, ip_address
from typing import Sequence

from visions.types.ip_address import IPAddress
from visions.types.string import String


@IPAddress.register_relationship(String, Sequence)
def string_is_ip_address(sequence: Sequence, state: dict) -> bool:
    try:
        _ = list(string_to_ip_address(sequence, state))
        return True
    except (ValueError, TypeError, AttributeError):
        return False


@IPAddress.register_transformer(String, Sequence)
def string_to_ip_address(sequence: Sequence, state: dict) -> Sequence:
    return tuple(map(ip_address, sequence))


@IPAddress.contains_op.register
def ip_address_contains(sequence: Sequence, state: dict) -> bool:
    return all(isinstance(x, _BaseAddress) for x in sequence)



# import pandas as pd
#
# from visions.relations.relations import InferenceRelation
# from visions.relations.relations_utils import values_are_consecutive
# from visions.types.ordinal import to_ordinal
#
#
# def is_ordinal_str(s: pd.Series, state: dict) -> bool:
#     if s.str.len().max() == 1:
#         unique_values = list(s[s.notna()].str.lower().unique())
#         return "a" in unique_values and values_are_consecutive(
#             list(map(ord, unique_values))
#         )
#     else:
#         return False
#
#
# def string_to_ordinal() -> InferenceRelation:
#     from visions.types import String
#
#     return InferenceRelation(
#         related_type=String,
#         relationship=is_ordinal_str,
#         transformer=to_ordinal,
#     )



# import pandas as pd
#
# from visions.relations.relations import InferenceRelation
# from visions.relations.relations_utils import values_are_consecutive
# from visions.types.ordinal import to_ordinal
# from visions.utils import func_nullable_series_contains
#
#
# @func_nullable_series_contains
# def is_ordinal_cat(series: pd.Series, state: dict) -> bool:
#     initial_element = "a"
#     s = series.astype(str)
#     if s.str.len().max() == 1:
#         distinct_values = list(s.str.lower().unique())
#         return initial_element in distinct_values and values_are_consecutive(
#             list(map(ord, distinct_values))
#         )
#     else:
#         return False
#
#
# def categorical_to_ordinal() -> InferenceRelation:
#     from visions.types import Categorical
#
#     return InferenceRelation(
#         Categorical, relationship=is_ordinal_cat, transformer=to_ordinal
#     )



from typing import Any, Sequence

from multimethod import multimethod

from visions.relations import IdentityRelation, InferenceRelation, TypeRelation
from visions.types.date_time import DateTime
from visions.types.object import Object
from visions.types.type import VisionsBaseType


class Date(VisionsBaseType):
    """**Date** implementation of :class:`visions.types.type.VisionsBaseType`.
    All values are should be datetime.date or missing

    Examples:
        >>> import datetime
        >>> import visions
        >>> x = [datetime.date(2017, 3, 5), datetime.date(2019, 12, 4)]
        >>> x in visions.Date
        True
    """

    @staticmethod
    def get_relations() -> Sequence[TypeRelation]:
        relations = [
            IdentityRelation(Object),
            InferenceRelation(DateTime),
        ]
        return relations

    @staticmethod
    @multimethod
    def contains_op(item: Any, state: dict) -> bool:
        pass



import uuid

import pandas as pd

from visions.backends.pandas.series_utils import (
    isinstance_attrs,
    series_handle_nulls,
    series_not_empty,
)
from visions.backends.pandas.test_utils import coercion_true_test
from visions.backends.shared.parallelization_engines import pandas_apply
from visions.types.string import String
from visions.types.uuid import UUID


@UUID.register_relationship(String, pd.Series)
def uuid_is_string(series: pd.Series, state: dict) -> bool:
    def f(s):
        return pandas_apply(s, uuid.UUID)

    return coercion_true_test(f)(series)


@UUID.register_transformer(String, pd.Series)
def uuid_to_string(series: pd.Series, state: dict) -> pd.Series:
    return pandas_apply(series, uuid.UUID)


@UUID.contains_op.register
@series_not_empty
@series_handle_nulls
def uuid_contains(series: pd.Series, state: dict) -> bool:
    return isinstance_attrs(series, uuid.UUID, ["time_low", "hex"])



from typing import Callable, List, Type

import attr
import pandas as pd

from visions.backends.shared.utilities import has_import


@attr.s
class Engine:
    name = attr.ib()

    @classmethod
    def setup(cls, *args, **kwargs) -> None:
        raise NotImplementedError("No setup defined for generic engine")

    @staticmethod
    def apply(series: pd.Series) -> Callable[[Callable], pd.Series]:
        raise NotImplementedError("No apply defined for generic engine")


class PandasEngine(Engine):
    name = "pandas"
    _is_setup = True

    @classmethod
    def setup(cls, *args, **kwargs) -> None:
        pass

    @staticmethod
    def apply(series: pd.Series) -> Callable[[Callable], pd.Series]:
        return series.apply


class SwifterEngine(Engine):
    name = "swifter"
    _is_setup = False

    @classmethod
    def setup(cls, *args, **kwargs) -> None:
        if cls._is_setup:
            return

        import swifter

        cls._is_setup = True

    @staticmethod
    def apply(series: pd.Series) -> Callable[[Callable], pd.Series]:
        return series.swifter.apply


_PANDAS_ENGINES = [PandasEngine, SwifterEngine]


class EngineCollection:
    def __init__(self, engines: List[Type[Engine]]):
        self.engines = {engine.name: engine for engine in engines}

    def is_engine(self, name: str) -> bool:
        return name in self.engines

    def get(self, name: str) -> Type[Engine]:
        return self.engines[name]


class PandasApply:
    supported_engines = EngineCollection(
        [engine for engine in _PANDAS_ENGINES if hasattr(engine, "apply")]
    )
    _engine: Type[Engine] = PandasEngine

    @property
    def engine(self) -> Type[Engine]:
        return self._engine

    @engine.setter
    def engine(self, value: str, *args, **kwargs) -> None:
        if not self.supported_engines.is_engine(value):
            raise ValueError(f"{value} is not a supported pandas apply engine")
        self._engine = self.supported_engines.get(value)
        self._engine.setup(*args, **kwargs)

    @property
    def apply(self) -> Callable[[pd.Series], Callable[[Callable], pd.Series]]:
        return self.engine.apply


class PandasHandler:
    def __init__(self):
        self.has_swifter = has_import("swifter")

        self.applier = PandasApply()
        self._set_default_apply_engine()

    def _set_default_apply_engine(self) -> None:
        if self.has_swifter:
            self.applier.engine = "swifter"


_pandas_handler = PandasHandler()


def pandas_apply(series: pd.Series, func: Callable) -> pd.Series:
    return _pandas_handler.applier.apply(series)(func)
